from collections import Counter
import os
import time
import sys

from astropy.coordinates import SkyCoord
from astropy.io import fits
from astropy.io.votable import parse_single_table
from astropy.table import Table
from astropy.time import Time
import astropy.units as u
import numpy as np
import pandas as pd

import ssos.filt as filt
from ssos.opt import extract_cutouts
from ssos.opt import compute_aperture_magnitudes
from ssos.opt import crossmatch_skybot
from ssos.utils import create_target_dir
from ssos.utils import init_argparse
from ssos.utils import init_logger
from ssos.utils import query_skybot
from ssos.utils import unpack_header_kw



FILTER_STEPS = {
        'FILTER_DETEC':        filt.detections,
        'FILTER_PM':           filt.proper_motion,
        'FILTER_MOTION':       filt.linear_motion,
        'FILTER_PIXEL':        filt.pixel,
        'FILTER_TRAIL':        filt.constant_trail,
        'FILTER_BRIGHT_SOURCES': filt.bright_sources_catalog
        }


ANALYSIS_STEPS = {
    'CROSSMATCH_SKYBOT': crossmatch_skybot,
    'EXTRACT_CUTOUTS':   extract_cutouts,
    'FIXED_APER_MAGS':   compute_aperture_magnitudes
    }


class Pipeline:
    ''' Class to handle the pipeline execution '''

    def __init__(self):

        # Preparations: Handling of command line arguments
        self.args = init_argparse()


        # Assert that images are found and contain the required header keywords
        self.images = [os.path.join(os.path.abspath(self.args.fields[0]), image) for image in
                       os.listdir(os.path.abspath(self.args.fields[0])) if image.endswith('.fits')]
        assert len(self.images) > 0, 'No images found in %s! Ensure that they have a .fits'\
                                     ' extension' % os.path.abspath(self.args.fields[0])

        self.target_dir, self.paths = create_target_dir(self.args)
        self.log, self.log_file, self.start_time = init_logger(self.args, self.paths['logs'])
        self.log.info('\n\t--- The ssos Pipeline ---\t\t--- {:s} ---\n\n'
                      .format(time.strftime('%Y/%m/%d %H:%M:%S', self.start_time)))


        # Reading and checking the settings
        self.settings = self._set_settings()
        self.settings = self._check_settings(self.settings)

        self._print_field_info()

        self.steps = [param for param in FILTER_STEPS.keys() if self.settings[param]]
        self.analysis_steps = [param for param in ANALYSIS_STEPS.keys() if self.settings[param]]
        self.added_proper_motion = False
        self.added_SExtractor_data = False

        # Little hack for fixed apertuer magnitudes "without" cutout extraction
        if not self.settings['EXTRACT_CUTOUTS'] and self.settings['FIXED_APER_MAGS']:
            self.log.debug('FIXED_APER_MAGS is True but EXTRACT_CUTOUTS is False. Will'\
                           ' delete cutouts upon finishing.')
            self.paths['cutouts'] = self.paths['tmp']
            self.analysis_steps[0:0] = ['EXTRACT_CUTOUTS']


    def _set_settings(self):
        '''
        Set pipeline settings

        input
        ------
        args - command line arguments parser
        log - logging instance

        return
        ------
        settings - dict, dictionary containing PARAMETER: VALUE pairs
        '''

        settings = {}

        try:
            if self.args.set_up:
                set_up = open(self.args.set_up[0], 'r')
            else:
                set_up = open('default.ssos', 'r')

        except IOError:
            self.log.debug('No config file provided in CWD or via -c flag.\
                            Using default settings.\n')
            try:
                set_up = open(os.path.join(os.path.dirname(__file__),
                                           'default.ssos'))
            except IOError:
                raise PipelineSettingsException('No configuration file provided '\
                       'and none found in package directory.')

        with set_up:
            for line in set_up:
                if line == '\n' or line[0] == '#':
                    continue

                parameter, value = line.split()[:2]
                settings[parameter] = value

        # Override values from command line
        for arg in vars(self.args):
            if arg in settings.keys() and vars(self.args)[arg] != None:
                settings[arg] = vars(self.args)[arg]

        # Expand $HOME to home directory
        settings = {k: v.replace('$HOME', os.path.expanduser('~')) for k, v in settings.items()}

        # Filter setting finished. Write them to the logfile
        with open(os.path.join(self.paths['logs'], self.log_file), 'a') as logfile:

            for setting, value in settings.items():
                line = setting.ljust(20) + str(value) + '\n'
                self.log.debug(line)
                logfile.write(line)

            logfile.write('\n')
            self.log.debug('\n')

        return settings


    def _check_settings(self, settings):
        '''
        Converts parameter values to expected formats

        input
        -----
        settings - dict, dictionary containing PARAMETER: VALUE pairs

        return
        -----
        settings - dict, dictionary containing PARAMETER: VALUE pairs
        '''

        # Unpack provided DETECTIONS and SCI extension(s) into list
        for param in ['SCI_EXTENSION', 'DETECTIONS']:
            try:
                settings[param] = [int(character) for character in settings[param].split(',')]
            except ValueError:
                if param == 'SCI_EXTENSION' and settings[param] == 'All':
                    settings[param] = False # treat it as 'None provided'
                else:
                    raise PipelineSettingsException('%s value invalid' % param)
        if self.settings['SCI_EXTENSION']:
            self.sci_ext = self.settings['SCI_EXTENSION'][0]
        else:
            self.sci_ext = 0

        # Check one image header for keyword presence
        with fits.open(self.images[0]) as exposure:
            kws = ['RA', 'DEC', 'OBJECT', 'DATE-OBS', 'FILTER', 'EXPTIME']
            for kw in kws:
                if not unpack_header_kw(exposure, self.settings[kw], self.sci_ext):
                    raise PipelineSettingsException('Could not find keyword %s in FITS header.'\
                                                    ' Is the SCI_EXTENSION correct?' % self.settings[kw])
                if kw == 'DATE-OBS':
                    # Find format of DATE-OBS keyword
                    try:
                        Time(unpack_header_kw(exposure, kw, try_first=self.sci_ext), format='isot')
                        self.date_obs_fmt = 'isot'
                    except ValueError:
                        self.date_obs_fmt = 'mjd'


        # Check that config files exist
        for file in ['SEX_CONFIG', 'SEX_PARAMS', 'SEX_NNW', 'SEX_FILTER',
                     'SCAMP_CONFIG', 'SWARP_CONFIG']:
            if not os.path.isfile(settings[file]):
                raise PipelineSettingsException('Could not find %s in %s' %
                                                (file, settings[file]) )

        # Check if weight images are provided and exist
        if not settings['WEIGHT_IMAGES'].upper() == 'FALSE':
            if not os.path.isdir(settings['WEIGHT_IMAGES']):
                raise PipelineSettingsException('Could not find weight images directory %s.' %
                                                 settings['WEIGHT_IMAGES'])
        else:
            settings['WEIGHT_IMAGES'] = False

        # Convert filter strings to booleans
        for param in ['REMOVE_REF_SOURCES', 'FILTER_DETEC', 'FILTER_PM', 'FILTER_PIXEL', 'FILTER_MOTION',
                      'IDENTIFY_OUTLIER', 'FILTER_TRAIL',
                      'FILTER_BRIGHT_SOURCES', 'CROSSMATCH_SKYBOT', 'EXTRACT_CUTOUTS',
                      'FIXED_APER_MAGS']:

            if settings[param].upper() == 'TRUE':
                settings[param] = True
            elif settings[param].upper() == 'FALSE':
                settings[param] = False
            else:
                raise PipelineSettingsException('Could not evaluate %s value, has to be True or\
                                                 False' % param)

        if settings['FILTER_MOTION']:
            if not settings['FILTER_DETEC'] or\
              (not 1 in settings['DETECTIONS'] and not 2 in settings['DETECTIONS']):
                raise PipelineSettingsException('When FILTER_MOTION is True, DETECTIONS needs\
                                                 to contain "1,2".')

        # Convert numeric values to float
        for param in ['PM_LOW', 'PM_UP', 'PM_SNR', 'DELTA_PIXEL', 'OUTLIER_THRESHOLD',
                      'R_SQU_M', 'RATIO', 'DISTANCE', 'CROSSMATCH_RADIUS',
                      'CUTOUT_SIZE']:

            try:
                settings[param] = float(settings[param])
            except ValueError:
                raise PipelineSettingsException('Could not convert %s value to float'
                                                % param)
            if param == ['R_SQU_M']:
                assert 0 <= settings[param] <= 1, 'The %s parameter has to be in range [0 - 1]' % param

        if settings['FILTER_BRIGHT_SOURCES']:
            # Evaluate bright-sources catalogue path
            if settings['BRIGHT_SOURCES_CAT'] == 'REFCAT':
                # get the filename and column names from the scamp config file
                with open(settings['SCAMP_CONFIG'], 'r') as file:
                    for line in file:
                        if 'REFOUT_CATPATH' in line:
                            refout_catpath = line.split()[1]
                        if 'ASTREF_CATALOG' in line:
                            astref_catalog = line.split()[1]
                if self.args.ASTREF_CATALOG: # if overwritten via command line
                    astref_catalog = self.args.ASTREF_CATALOG

                settings['BRIGHT_SOURCES_CAT'] = [refout_catpath, astref_catalog]

        return settings


    def _print_field_info(self):
        ''' Prints RA, DEC, and OBJECT keywords to log '''
        with fits.open(self.images[0]) as exposure:
            ra, dec, object = unpack_header_kw(exposure, ['RA', 'DEC', 'OBJECT'], try_first=self.sci_ext)

        ecli_lat = SkyCoord(ra, dec, frame='icrs', unit='deg').barycentrictrueecliptic.lat.deg

        self.log.info('\t|\t'.join(['%i Exposures' % len(self.images),
                                    '%s' % object,
                                    '%.2fdeg Ecliptic Latitude\n\n' % ecli_lat]))

        # Get number of known SSOs in FoV from SkyBoT
        if self.settings['CROSSMATCH_SKYBOT']:
            self.log.info('Querying SkyBoT for known SSOs in FoV..\n')

            if not self.args.skybot and \
               os.path.isfile(os.path.join(self.paths['skybot'], 'skybot_all.csv')):

               self.skybot = pd.read_csv(os.path.join(self.paths['skybot'], 'skybot_all.csv'))

            else:
                self.skybot = pd.DataFrame()

                for img in self.images:
                    with fits.open(img) as exp:
                        ra, dec, date_obs, texp = unpack_header_kw(exp, [self.settings['RA'], self.settings['DEC'],
                                                                         self.settings['DATE-OBS'], self.settings['EXPTIME']], self.sci_ext)

                        mid_epoch = (Time(date_obs, format=self.date_obs_fmt) + float(texp) / 2 * u.second).isot

                        if self.settings['FOV_DIMENSIONS'] == '0x0':

                            # Determine exposure FoV
                            naxis1, naxis2, cdelt1, cdelt2 = unpack_header_kw(exp, ['NAXIS1', 'NAXIS2', 'CDELT1', 'CDELT2'], self.sci_ext)

                            if not cdelt1 or not cdelt2:
                                cd11, cd12, cd21, cd22 = unpack_header_kw(exp, ['CD1_1', 'CD1_2', 'CD2_1', 'CD2_2'], self.sci_ext)
                                dra  = naxis1 * abs(cd11) + naxis2 * abs(cd12)
                                ddec = naxis1 * abs(cd21) + naxis2 * abs(cd22)

                            else:
                                dra = naxis1 * cdelt1
                                ddec = naxis2 * cdelt2

                            # Ensure coverage by adding CROSSMATCH_RADIUS
                            fov = '{:.1f}x{:.1f}'.format(round(dra  + self.settings['CROSSMATCH_RADIUS'] / 60, 1),
                                                         round(ddec + self.settings['CROSSMATCH_RADIUS'] / 60, 1))

                        else:
                            fov = self.settings['FOV_DIMENSIONS']

                        result = query_skybot(mid_epoch, ra, dec, fov,
                                              self.settings['OBSERVATORY_CODE'])
                        if not result.empty:
                            self.skybot = self.skybot.append(result)

                if not self.skybot.empty:
                    self.skybot.to_csv(os.path.join(self.paths['skybot'], 'skybot_all.csv'), index=False)

            if not self.skybot.empty:
                # Print gimmicky SkyBoT info
                bins = np.arange(np.floor(min(self.skybot.Mv)),
                                 np.ceil(max(self.skybot.Mv)), 0.5)

                counts = pd.cut(self.skybot['Mv'], bins).value_counts()
                counts.sort_index(inplace=True)

                magnitudes = '  '.join([str(i) for i in range(int(min(bins)),
                                                             int(np.ceil(max(bins))) + 1
                                                             )])
                # construct bar chart top-down
                bars = []
                bar_height = 7

                for i in range(1, bar_height):
                    bar = ' '
                    for count in counts:
                        if count >= max(counts.values) * (bar_height-i)/bar_height:
                            bar += '##'
                        else:
                            bar += '  '
                    bars.append(bar)
                # bottom bar
                bar = ' '
                for count in counts:
                    if count > 0:
                        bar += '##'
                    else:
                        bar += '  '
                bars.append(bar)

                self.log.info('\nSkyBoT: %i SSOs with %i detections\n'
                              '\n%s\n%s\n%s\n%s\n%s\n%s\n%s\n %s\n%sMv\n'
                               % (len(set(self.skybot.Name)),
                                  len(self.skybot),
                                  *bars, magnitudes,
                                  ' ' * int(len(magnitudes) / 2)))
            else:
                self.log.info('No known SSOs returned by SkyBoT.\n')

    def _run_SExtractor_on_single_image(self, image, extension):
        '''
        Run SExtractor on individual images.

        input
        ------
        image - str, absolute path to image file
        extension - int, index of FITS file extension to be SExtracted

        return
        ------
        cat, str - absolute path to SExtractor output catalog
        date_obs, str - observation epoch from exposure
        '''

        # Name of output catalog
        cat = os.path.join(self.paths['cats'],
                           os.path.splitext(os.path.basename(image))[0])

        # Select extension or run whole image
        if extension is not False:
            image_ext = image + '[%i]' % extension
            cat += '_%i.cat' % extension
            ext = extension

        else:
            image_ext = image
            cat += '.cat'
            ext = 1

        if not self.args.sex and os.path.isfile(cat):
            self.log.debug('SExtractor catalog %s already exists! Skipping this sextraction..\n' % cat)

            with fits.open(image) as exposure:
                date_obs = unpack_header_kw(exposure, self.settings['DATE-OBS'], self.sci_ext)

            return cat, date_obs

        sex_args = {

            'file': image_ext,
            'config': self.settings['SEX_CONFIG'],
            'overwrite_params': { # Arguments to initialize Astromatic class
                    'CATALOG_NAME': cat,
                    'PARAMETERS_NAME': self.settings['SEX_PARAMS'],
                    'FILTER_NAME': self.settings['SEX_FILTER'],
                    'STARNNW_NAME': self.settings['SEX_NNW'],
            },
        }

        if self.settings['WEIGHT_IMAGES']:
            sex_args['overwrite_params']['WEIGHT_IMAGE'] = 'MAP_WEIGHT'

            if extension is not False:
                weight_suffix = '_%i.weight' % extension
            else:
                weight_suffix = '.weight'

            sex_args['overwrite_params']['WEIGHT_IMAGE'] = os.path.join(self.settings['WEIGHT_IMAGES'],
                                                           os.path.basename(image).replace(
                                                           '.fits', weight_suffix))

        if self.log.level <= 10:  # if we're at DEBUG log level, print SExtractor output
            sex_args['overwrite_params']['VERBOSE_TYPE'] = 'NORMAL'

        # ------
        # Exectue SExtractor
        cmd = ' '.join(['sex', sex_args['file'], '-c', sex_args['config']])
        for param, value in sex_args['overwrite_params'].items():
            cmd += ' '.join([' -' + param, value])

        self.log.debug('\nExecuting SExtractor command:\n%s\n\n' % cmd)
        os.system(cmd)

        # ------
        with fits.open(image) as exposure:
            # Make .ahead file with the EPOCH in MJD for SCAMP
            # Following http://www.astromatic.net/forum/showthread.php?tid=501
            date_obs = unpack_header_kw(exposure, self.settings['DATE-OBS'], self.sci_ext)
            mjd = Time(date_obs, format=self.date_obs_fmt).mjd

            with open(os.path.splitext(cat)[0] + '.ahead', 'w+') as file:
                for hdu in exposure:
                    file.write('MJD-OBS = %.6f\nEND\n' % mjd)
        return cat, date_obs


    def run_SExtractor(self):
        ''' Wrapper for SExtractor run.  Adds progress bar.'''

        self.SExtractor_catalogues = []

        observation_epochs = []
        # Add progress bar
        if not self.args.quiet:
            sys.stdout.write('[%s]' % (' ' * len(self.images)))
            sys.stdout.flush()
            sys.stdout.write('\b' * (len(self.images) + 1))

        for image in self.images:

            if self.settings['SCI_EXTENSION']:

                for extension in self.settings['SCI_EXTENSION']:
                    cat, date_obs = self._run_SExtractor_on_single_image(image, extension)
                    observation_epochs.append(date_obs)
                    self.SExtractor_catalogues.append(cat)

            else:
                cat, date_obs = self._run_SExtractor_on_single_image(image, False)
                observation_epochs.append(date_obs)
                self.SExtractor_catalogues.append(cat)

            if not self.args.quiet:
                sys.stdout.write('-')
                sys.stdout.flush()

        # Sort the SExtractor catalogues by their observation epochs, important for SCAMP
        self.SExtractor_catalogues = [cat for _, cat in sorted(zip(observation_epochs,
                                                               self.SExtractor_catalogues))]


    def run_SCAMP(self, crossid_radius=None, full_name='full.cat',
                  merged_name='merged.cat', keep_refcat=False,
                  solve_astronomy=True, pattern_matching=True,
                  adjust_SExtractor_and_aheader=False):
        ''' Run SCAMP on SExtractor catalogs '''

        # ------
        # Set-up SCAMP run
        self.merged_cat = os.path.join(self.paths['cats'], merged_name)
        self.full_cat = os.path.join(self.paths['cats'], full_name)

        scamp_args = {
            'files':  self.SExtractor_catalogues,
            'config': self.settings['SCAMP_CONFIG'],

            'overwrite_params': {
                'MERGEDOUTCAT_NAME': self.merged_cat,
                'FULLOUTCAT_NAME':   self.full_cat,
                },
            }

        # SCAMP appends an index for the Field Group to the output catalogues
        self.merged_cat = self.merged_cat.replace('.cat', '_1.cat')
        self.full_cat = self.full_cat.replace('.cat', '_1.cat')

        if self.args.ASTREF_CATALOG is not None:
            scamp_args['overwrite_params']['ASTREF_CATALOG'] = self.args.ASTREF_CATALOG

        if self.settings['FILTER_BRIGHT_SOURCES'] and \
           self.settings['BRIGHT_SOURCES_CAT'] == 'REFCAT':
            scamp_args['overwrite_params']['SAVE_REFCATALOG'] = 'Y'

        if not solve_astronomy:
            scamp_args['overwrite_params']['SOLVE_ASTROM'] = 'N'
            scamp_args['overwrite_params']['INCLUDE_ASTREFCATALOG'] = 'Y'

        if not pattern_matching:
            scamp_args['overwrite_params']['MATCH'] = 'N'

        if crossid_radius is not None:
            scamp_args['overwrite_params']['CROSSID_RADIUS'] = str(crossid_radius)

        if not self.args.scamp and os.path.isfile(self.merged_cat) and os.path.isfile(self.full_cat):
            if not pattern_matching: # suppress this message in case SCAMP is run twice
                self.log.info('\nReading SCAMP catalogues from file..\n')
            scamp_from_file = True

        else:
            self.log.info('\nRunning SCAMP..\t')

            # ------
            # Exectue SCAMP
            catalogs = ' '.join(scamp_args['files'])
            cmd = ' '.join(['scamp', catalogs, '-c', scamp_args['config']])
            for param, value in scamp_args['overwrite_params'].items():
                cmd += ' '.join([' -' + param, value])

            self.log.debug('\nExecuting SCAMP command:\n%s\n' % cmd)
            os.system(cmd)
            scamp_from_file = False

        # Create the full catalog as pipeline property
        with fits.open(self.full_cat) as full:
            data = Table(full[2].data)

        self.sources = data.to_pandas()

        if not keep_refcat:
            # Catalogue number 0 are the reference sources
            self.sources = self.sources[self.sources['CATALOG_NUMBER'] != 0]

        if (self.args.scamp or not scamp_from_file) \
            and adjust_SExtractor_and_aheader:
            self.adjust_SExtractor_catalogues()
            self.adjust_ahead_files()

        # Ensure that EPOCHs are in order
        self.sources = self.sources.sort_values(['SOURCE_NUMBER', 'EPOCH'],
                                                ascending=[True, True])
        self.sources = self.sources.reset_index()

        # Add flag to dataframe
        self.sources['FLAGS_SSOS'] = 0


    def adjust_SExtractor_catalogues(self):
        # Flags source detections which were associated to reference sources
        # These will subsequently be ignored by SCAMP
        self.log.info('Adjusting SExtractor catalogues for pattern matching..\n')
        REMOVE_DETECTIONS = 2 # 1 image + reference, lower limit

        detections = Counter(self.sources.SOURCE_NUMBER) # this includes references
        flag_count = [source for source, count in detections.items() if count >= REMOVE_DETECTIONS]

        # only remove sources which were associated to a reference
        flag = []
        for source_number in flag_count:
            if 0 in self.sources[self.sources.SOURCE_NUMBER == source_number]['CATALOG_NUMBER'].values:
                flag.append(source_number)

        self.sources = self.sources[self.sources.SOURCE_NUMBER.isin(flag)]
        self.sources = self.sources[self.sources.CATALOG_NUMBER != 0]

        for cat_number, catalogue in self.sources.groupby('CATALOG_NUMBER'):
            with fits.open(self.SExtractor_catalogues[cat_number-1]) as cat:

                for extension, detections in catalogue.groupby('EXTENSION'):
                    extension *= 2 # extensions are HEADER and DATA

                    xwin_img = cat[extension].data.field('XWIN_IMAGE')
                    detections_to_flag = np.in1d(xwin_img, detections.X_IMAGE)
                    cat[extension].data['FLAGS'][detections_to_flag] = 128

                cat.writeto(self.SExtractor_catalogues[cat_number-1], overwrite=True)

    def adjust_ahead_files(self):
        self.log.info('Copying astrometric solution to aheader files..\n')
        aheaders = [cat.replace('.cat', '.ahead') for cat in self.SExtractor_catalogues]

        for ahead in aheaders:
            with open(ahead, 'r') as ahead_file:
                mjd_obs = ahead_file.readline()

            with open(ahead, 'w') as ahead_file:
                with open(ahead.replace('.ahead', '.head'), 'r') as head_file:
                    for line in head_file:
                        if 'END' in line:
                            line = mjd_obs + line
                        ahead_file.write(line)


    def number_of_sources(self):
        return len(set(self.sources.SOURCE_NUMBER))


    def check_remaining(func):
        ''' Wrapper to check if source candidates remain after filter step'''
        def wrapped_func(*args, **kwargs):

            result = func(*args, **kwargs)

            if not args[0].number_of_sources() > 0:
                raise NoSourcesRemainingException('\nNo source candidates left after %s.'
                                                   ' Stopping pipeline.\n' % args[1])
                sys.exit()

            return result
        return wrapped_func


    @check_remaining
    def execute_filter(self, step_name):
        '''
        Calls the proper filter function

        input
        ------
        step_name - str, name of filter step

        return
        ------
        sources - pd.DataFrame, remaining source candidates
        '''

        # The proper motion filter has to access the merged catalogue first
        if step_name == 'FILTER_PM':
            self.add_proper_motion()

        # The trail filters have to access the SExtractor catalogues first
        if step_name in ['FILTER_PIXEL', 'FILTER_TRAIL', 'FILTER_T_DIST']:
            self.add_SExtractor_data()

        self.sources = FILTER_STEPS[step_name](self.sources, self.settings)


    def add_proper_motion(self):
        '''
        Adds proper motion values from merged to full catalog.  Has to be called
        either before filtering the proper motion or at the end of the pipeline.
        '''
        with fits.open(self.merged_cat) as merged:
            data = Table(merged[2].data)

        proper_motion_columns = ['PMALPHA_J2000', 'PMALPHAERR_J2000',
                                 'PMDELTA_J2000', 'PMDELTAERR_J2000']

        merged = data[['SOURCE_NUMBER'] + proper_motion_columns].to_pandas()

        # Add proper motions column from merged catalog to source database
        self.sources = self.sources.merge(merged, on='SOURCE_NUMBER')

        for column in proper_motion_columns:
            self.sources[column] = self.sources[column].apply(lambda x: x / 8.75e6)

        # Add proper motion column.  Factor 8.75e6 converts to "/h
        self.sources['PM'] = self.sources.apply(lambda x:
                               np.sqrt(x['PMALPHA_J2000']**2 + x['PMDELTA_J2000']**2),
                               axis=1)
        self.sources['PMERR'] = self.sources.apply(lambda x:
                               np.sqrt(x['PMALPHAERR_J2000']**2 + x['PMDELTAERR_J2000']**2),
                               axis=1)

        self.added_proper_motion = True


    def add_SExtractor_data(self):
        '''
        Adds the trail morphology parameters from the SExtractor catalogues to
        the source database.  Has to be called either before filtering on trail
        morphology or at the end of the pipeline.
        '''

        # Add SExtractor data
        morphometry_columns = ['AWIN_IMAGE', 'BWIN_IMAGE', 'THETAWIN_IMAGE',
                               'XWIN_IMAGE', 'YWIN_IMAGE', 'ERRAWIN_IMAGE',
                               'ERRBWIN_IMAGE', 'ERRTHETAWIN_IMAGE', 'FLUX_AUTO',
                               'FLUXERR_AUTO']
        self.sources['XWIN_IMAGE'] = self.sources['X_IMAGE']

        sextractor_data = pd.DataFrame(index=[], columns=morphometry_columns)


        for catalogue in self.SExtractor_catalogues:
            with fits.open(catalogue) as cat:
                for j, header in enumerate(cat):
                    if j % 2 == 0 and j != 0:
                        data = Table(header.data)[:][morphometry_columns]
                        sextractor_data = sextractor_data.append(data.to_pandas(),
                                                     ignore_index=True, sort=True)

        self.sources = pd.merge(self.sources, sextractor_data[morphometry_columns],
                                on='XWIN_IMAGE', how='left')
        self.added_SExtractor_data = True


    def add_image_metadata(self):

        # Derive image filename from SExtractor catalogue
        for cat_number, group in self.sources.groupby('CATALOG_NUMBER'):

            if self.settings['SCI_EXTENSION']:
                image_filename = '_'.join(os.path.splitext(
                                            os.path.basename(self.SExtractor_catalogues[cat_number-1])
                                                          )[0].split('_')[:-1]) + '.fits'
            else:
                image_filename = os.path.basename(self.SExtractor_catalogues[cat_number-1]).replace('.cat', '.fits')

            self.sources.loc[group.index, 'IMAGE_FILENAME'] = image_filename

        # Add image metadata. Have to use the correct header extension
        for image_filename, group in self.sources.groupby(['IMAGE_FILENAME']):

            extension = group.EXTENSION.values[0]
            # Add exposure keywords
            with fits.open(os.path.join(self.paths['images'], image_filename)) as exposure:
                for prop in ['OBJECT', 'DATE-OBS', 'FILTER', 'EXPTIME', 'RA_IMAGE', 'DEC_IMAGE']:
                    self.sources.loc[group.index, prop] = date_obs = unpack_header_kw(exposure, self.settings[prop.split('_')[0]], extension)

        self.sources['MID_EXPOSURE_MJD'] = self.sources.apply(lambda x: ( Time(x['DATE-OBS'], format=self.date_obs_fmt) +
                                                                       float(x['EXPTIME']) / 2 * u.second).mjd, axis=1)


    def execute_analysis(self, step_name):
        '''
        Calls the proper analysis function

        input
        ------
        step_name - str, name of analysis step
        '''
        self.sources = ANALYSIS_STEPS[step_name](self.sources, self.settings, self.log,
                                      self.paths, self.args)


    def save_and_cleanup(self):
        ''' Clean up the final database.  Rename columns, remove unnecessary columns. '''

        self.sources.drop(columns=['ASTR_INSTRUM', 'PHOT_INSTRUM'], inplace=True)

        self.sources.rename(index=str, columns={'ALPHA_J2000': 'RA', 'DELTA_J2000': 'DEC', 'PMALPHA_J2000': 'PMRA',
                                                'PMALPHAERR_J2000': 'PMRA_ERR', 'PMDELTA_J2000': 'PMDEC',
                                                'PMDELTAERR_J2000': 'PMDEC_ERR', 'SKYBOT_PMALPHA': 'SKYBOT_PMRA',
                                                'SKYBOT_PMDELTA': 'SKYBOT_PMDEC', 'SKYBOT_ALPHA': 'SKYBOT_RA',
                                                'SKYBOT_DELTA': 'SKYBOT_DEC', 'FLUX_AUTO': 'FLUX',
                                                'FLUXERR_AUTO': 'FLUXERR'}, inplace=True)

        # Rearrange columns
        ordered_columns = ['SOURCE_NUMBER', 'CATALOG_NUMBER', 'RA', 'DEC', 'EPOCH', 'MAG', 'MAGERR', 'FLUX', 'FLUXERR',
                           'PM', 'PMERR', 'PMRA',  'PMRA_ERR', 'PMDEC', 'PMDEC_ERR', 'MID_EXPOSURE_MJD', 'DATE-OBS',
                           'EXPTIME', 'OBJECT', 'FILTER', 'RA_IMAGE', 'DEC_IMAGE', 'IMAGE_FILENAME',
                           'EXTENSION', 'XWIN_IMAGE', 'YWIN_IMAGE', 'AWIN_IMAGE', 'ERRAWIN_IMAGE', 'BWIN_IMAGE',
                           'ERRBWIN_IMAGE', 'THETAWIN_IMAGE', 'ERRTHETAWIN_IMAGE', 'ERRA_WORLD', 'ERRB_WORLD',
                           'ERRTHETA_WORLD', 'FLAGS_EXTRACTION', 'FLAGS_SCAMP', 'FLAGS_IMA', 'FLAGS_SSOS']

        if self.settings['FIXED_APER_MAGS']:
            ordered_columns[7:7] = ['MAG_CI', 'MAGERR_CI', 'FLUX_CI', 'FLUXERR_CI']

        if self.settings['CROSSMATCH_SKYBOT']:
            ordered_columns[22:22] = ['SKYBOT_NUMBER', 'SKYBOT_NAME', 'SKYBOT_CLASS', 'SKYBOT_MAG',
                                      'SKYBOT_RA', 'SKYBOT_DEC', 'SKYBOT_PMRA', 'SKYBOT_PMDEC']

        self.sources = self.sources[ordered_columns]

        # Save final database
        output = Table.from_pandas(self.sources)
        output_filename = os.path.join(self.paths['cats'], 'ssos_{:s}.csv'.format(time.strftime('%Y%m%d%H%M%S', self.start_time)))
        output.write(output_filename, format='csv', overwrite=True)

        os.system('rm -r %s' % self.paths['tmp'])
        self.run_time = time.time() - time.mktime(self.start_time)


class PipelineSettingsException(Exception):
    pass


class NoSourcesRemainingException(Exception):
    pass



import argparse
import logging
import os
import sys
import time
import warnings

from astropy.coordinates import SkyCoord
import astropy.units as u
import pandas as pd


def init_argparse():
    '''
    Executes the setup of the arguments parser for the command-line api

    return
    ------
    args - command line argument parser
    '''

    parser = argparse.ArgumentParser(description='Pipeline to search for Solar System\
                                                  objects in wide-field imaging surveys')

    parser.add_argument('fields', nargs='+', help='Path to directory of field to be searched')

    parser.add_argument('-c', '--config', action='store', dest='set_up', nargs=1,
                        help='Path to configuration file')

    parser.add_argument('-d', '--default', action='store_true',
                        help='Copy default setup files to CWD')

    parser.add_argument('-i', '--inspect', action='store_true',
                        help='Launch visual inspection mode on SSO candidates sample')

    parser.add_argument('-l', '--log', action='store', dest='log', nargs=1,
                        help='Set the logging level. Valid arguments are DEBUG, INFO, WARNING, ERROR,\
                        CRITICAl.', default=['INFO'])

    parser.add_argument('-q', '--quiet', action='store_true',
                        help='Suppress logging to console')

    parser.add_argument('-t', '--target', action='store', dest='target', nargs=1,
                        help='Target directory to save fits files. If no target given, writing to CWD')

    parser.add_argument('--sex', action='store_true',
                        help='Force SExtractor runs')

    parser.add_argument('--scamp', action='store_true',
                        help='Force SCAMP runs')

    parser.add_argument('--swarp', action='store_true',
                        help='Force SWARP runs')

    parser.add_argument('--skybot', action='store_true',
                        help='Force SkyBoT query')

    group = parser.add_argument_group('Filter Settings')

    for prop in ['REMOVE_REF_SOURCES', 'FILTER_DETEC', 'FILTER_PM', 'FILTER_PIXEL', 'FILTER_MOTION', 'IDENTIFY_OUTLIER',
                 'FILTER_TRAIL', 'FILTER_BRIGHT_SOURCES', 'CROSSMATCH_SKYBOT',
                 'EXTRACT_CUTOUTS', 'FIXED_APER_MAGS']:

        group.add_argument('-' + prop, metavar='bool', action='store',
                           help='Override ' + prop + ' setting. Must be True or False.',
                           choices=['True', 'False'])


    for val in ['SCI_EXTENSION', 'WEIGHT_IMAGES', 'RA', 'DEC', 'OBJECT', 'DATE_OBS', 'FILTER', 'EXPTIME',
                'SEX_CONFIG', 'SEX_PARAMS', 'SEX_FILTER', 'SEX_NNW', 'SCAMP_CONFIG', 'ASTREF_CATALOG',
                'CROSSID_RADIUS', 'SWARP_CONFIG', 'DETECTIONS', 'PM_LOW', 'PM_UP', 'PM_SNR', 'DELTA_PIXEL',
                'OUTLIER_THRESHOLD', 'R_SQU_M', 'RATIO', 'BRIGHT_SOURCES_CAT', 'DISTANCE', 'MAG_LIMITS', 'CROSSMATCH_RADIUS',
                'CUTOUT_SIZE', 'REFERENCE_FILTER', 'OBSERVATORY_CODE', 'FOV_DIMENSIONS']:

        group.add_argument('-' + val, metavar='value', action='store',
                           help='Override ' + val + ' setting.')

    args = parser.parse_args()

    return args


def init_logger(args, log_dir):
    '''
    Initiate logging. By default, the log is written to file.
    The filename consists of the string 'sso_' followed by the current date and time.

    input
    ------
    args - command line argument parser
    log_dir - str, absolute path to log directory

    return
    ------
    log - logger instance
    log_file - str, absolute path to log file
    '''

    # See if logging level was set via argument. Else, default to INFO
    numeric_level = getattr(logging, args.log[0].upper(), None)

    if not isinstance(numeric_level, int):
        numeric_level = 20

    log = logging.getLogger(__name__)
    log.setLevel(numeric_level)

    start_time = time.struct_time(time.localtime())
    log_file = 'ssos_{:s}.log'.format(time.strftime('%Y%m%d%H%M%S', start_time))

    file_log = logging.FileHandler(os.path.join(log_dir, log_file))
    file_log.setLevel(numeric_level)
    file_log.terminator = '' # suppress newline characters
    log.addHandler(file_log)

    # If the quiet flag is set, logs to console are supressed
    if not args.quiet:
        stdout_log = logging.StreamHandler()
        stdout_log.setLevel(numeric_level)
        stdout_log.terminator = '' # suppress newline characters
        log.addHandler(stdout_log)

    def log_excepthook(execption_type, exception_value, traceback, logger=log):
        ''' Function to pipe tracebacks into the logfile '''
        log.error('Exception occurred during analysis: \n',
                  exc_info=(execption_type, exception_value, traceback))
        log.error('\n')

    sys.excepthook = log_excepthook
    return log, log_file, start_time


def create_target_dir(args):
    '''
    Creates the target directory file structure

    input
    ------
    args - command line argument parser

    return
    ------
    target_dir - str, absolute path to target directory
    paths - dict, dictionary containing the absolute paths of the subdirectories
    '''

    if not args.target:
        target_dir = os.getcwd()
    else:
        target_dir = os.path.abspath(args.target[0])

    paths = {
        'cats':    os.path.join(target_dir, 'cats'),
        'cutouts': os.path.join(target_dir, 'cutouts'),
        'logs':     os.path.join(target_dir, 'logs'),
        'images':  os.path.abspath(args.fields[0]),
        'skybot':  os.path.join(target_dir, 'skybot'),
        'tmp':  os.path.join(target_dir, 'tmp'),
        'weights':  os.path.join(target_dir, 'weights')
    }

    for _, path in paths.items():
        os.makedirs(path, exist_ok=True)

    return target_dir, paths


def query_skybot(epoch, ra, dec, fov, obs_code):
    '''
    Retrieves the SkyBoT query for the specified parameters

    input
    ------
    epoch - float, mid-exposure epoch in MJD
    ra - float, right ascension in degree
    dec - float, declination in degree
    fov - str, value of FOV_DIMENSIONS parameter
    obs_code - str, value of OBSERVATORY_CODE parameter

    return
    -------
    skybot - pd.DataFrame, skybot sources in the FoV. Empty dataframe if no sources
    '''

    obs = {
        '-ep': str(epoch), '-ra': str(ra),
        '-dec': str(dec),  '-bd': fov,
        '-mime': 'text',  '-loc': obs_code,
        '-output': 'obs'
    }

    url = 'http://vo.imcce.fr/webservices/skybot/skybotconesearch_query.php?'
    url = '&'.join([url, *['='.join([k, v]) for k, v in obs.items()]])

    skybot = pd.read_csv(url, skiprows=2, delimiter='|')

    if skybot.empty:
        print(skybot.columns)
        return skybot
    # Fix column names
    skybot = skybot.rename({col: col.replace(' ', '') for col in skybot.columns},
                           axis=1)

    # Convert RA and DEC columns to degree
    skybot['RA'] = skybot.apply(lambda x: SkyCoord(x['RA(h)'], x['DE(deg)'], frame='icrs', unit=(u.hourangle, u.deg)).ra.deg, axis=1)
    skybot['DEC'] = skybot.apply(lambda x: SkyCoord(x['RA(h)'], x['DE(deg)'], frame='icrs', unit=(u.hourangle, u.deg)).dec.deg, axis=1)
    skybot.drop(['RA(h)', 'DE(deg)'], inplace=True, axis=1)

    skybot['EPOCH'] = epoch

    return skybot


def unpack_header_kw(hdus, keywords, try_first=0):
    '''
    Looks up keyword values in FITS headers. Iterates over
    extensions until keyword is found, else returns False

    hdus - HDUList, exposure headers
    keywords - string or list, keywords to look up
    try_first - header index to look up first

    returns
    ----
    vals - single parameter value or list of parameter values
    '''
    vals = []

    kws = [keywords] if not isinstance(keywords, list) else keywords

    for kw in kws:
        for hdu_ind in set([try_first, *range(len(hdus))]):
            try:
                hdu = hdus[hdu_ind]
                vals.append(hdu.header[kw])
                break
            except (IndexError, KeyError):
                pass
        else:
            vals.append(False)

    if not isinstance(keywords, list):
        return vals[0]

    return vals


name = 'ssos'
__version__ = '1.2.3'


GREETING = '''
       *       +      ssos %s
   +        x
        '             https://ssos.readthedocs.io
*   +     _.::'   |
        :_.'    - o - ssos --default : Copy default settings to CWD
|                 |   ssos --help    : Print detailed help
O -     *        '    ssos [-PARAMETER] [-VALUE] path/to/images
|
        _..:'         max.mahlke(at)cab.inta-csic.es
  +   (_.:

        ''' % __version__



'''
    Python implementations of filter functions to detect SSOs in SExtractor and SCAMP catalogues

    Part of the ssos pipeline.
'''

import collections
import os

from astropy.io import fits
from astropy.coordinates import SkyCoord
from astropy.table import Table
import numpy as np
import pandas as pd
from statsmodels import robust
import statsmodels.api as sm



def detections(sources, settings):
    '''
        Remove sources based on number of detections

        input
        ------
        sources - pandas dataframe, table of source data
        detections - list, containing integers responding to number of detections

        return
        ------
        sources - pandas dataframe, sources data without the sources filtered in this step
    '''
    measurements = collections.Counter(sources['SOURCE_NUMBER'])
    sources_to_remove = {key for key, value in measurements.items()
                         if value in settings['DETECTIONS']}

    sources = sources[~sources['SOURCE_NUMBER'].isin(sources_to_remove)]
    return sources


def proper_motion(sources, settings):
    '''
    Remove sources based on the proper motion range and signal-to-noise

    input
    ------
    sources - pandas dataframe, SCAMP full catalog
    settings - dictionary, settings parameter:value pairs

    return
    ------
    sources - pandas dataframe, sources data without the sources filtered in this step
    '''
    pm_low = settings['PM_LOW']
    pm_up = settings['PM_UP']
    SN = settings['PM_SNR']

    # Filter sources for proper motion range
    sources = sources[(sources.PM >= pm_low) & (sources.PM <= pm_up)]

    # Filter by proper motion SN
    sources = sources[sources.PM / sources.PMERR >= SN]
    return sources


def pixel(sources, settings):
    '''
    Remove sources based on pixel coordinates

    input
    ------
    sources - pandas dataframe, SCAMP full catalog
    settings - dictionary, contains the specified filter parameters

    return
    ------
    sources - pandas dataframe, sources data without the sources filtered in this step
    '''
    delta_pixel = settings['DELTA_PIXEL']

    # For all sources, check if the largest XWIN_IMAGE !and! YWIN_IMAGE differences
    # are smaller than delta_pixel.
    for source_number, group in sources.copy().groupby('SOURCE_NUMBER'):

        x_img = group['XWIN_IMAGE']
        y_img = group['YWIN_IMAGE']

        if x_img.max() - x_img.min() < delta_pixel:
            if y_img.max() - y_img.min() < delta_pixel:
                sources = sources[sources.SOURCE_NUMBER != source_number]

    return sources


def motion_is_linear(epoch, coords, sigma, r_squared):
    '''
    Motion fitting function

    input
    ------
    epoch - np.array, epoch of source
    ra - pd.Series, right ascension of source
    dec - pd.Series, declination of source
    r_squared - float, minimum r_squared for good fit

    return
    ------
    True if motion is linear else False
    '''
    epoch = sm.add_constant(epoch)
    model = sm.WLS(coords, epoch, weights=1/sigma**2)
    results = model.fit()

    if results.rsquared >= r_squared:
        return True
    else:
        return False

def motion_is_linear_in_ecliptic_coordinates(epoch, ra, dec, r_squ):
    # Transform to ecliptic coordinate system
    ecliptic_coords = SkyCoord(ra, dec, frame='icrs', unit='deg').barycentrictrueecliptic
    if not motion_is_linear(epoch, ecliptic_coords.lon.deg, 1, r_squ):
        return False
    else:
        if not motion_is_linear(epoch, ecliptic_coords.lat.deg, 1, r_squ):
            return False
        else:
            return True


def linear_motion(sources, settings):
    '''
        Remove sources based on linearity of motion and the Median Absolute Deviation

        input
        ------
        sources - pandas dataframe, table of source data
        settings - dict, 'parameter': value

        return
        ------
        sources - pandas dataframe, sources data without the sources filtered in this step
    '''

    r_squ, out_thres, id_out = settings['R_SQU_M'], settings['OUTLIER_THRESHOLD'],\
                               settings['IDENTIFY_OUTLIER']

     # Lower limit of subgroup size
    min_detections = 3

    sources_to_remove = []

    for source_number, group in sources.copy().groupby('SOURCE_NUMBER'):
        # GROUP contains the detections of a single source
        epoch = group['EPOCH'].values
        ra = group['ALPHA_J2000'].values
        dec = group['DELTA_J2000'].values

        if id_out:
            # Detect outliers using the Median Absolute Deviation of the EPOCHS
            outlier_index = np.where(abs(np.diff(epoch)) > robust.mad(epoch) *
                                                           out_thres)[0]
            if len(outlier_index) == 0:  # no outlier in EPOCH. Check edges
                outlier_in_middle = np.where(abs(np.diff(epoch[:-1])) > robust.mad(epoch[:-1]) *
                                                                         out_thres)[0]
                if len(outlier_in_middle) == 0:
                    # Linear in right ascencsion?
                    if not motion_is_linear(epoch, ra, group['ERRA_WORLD'].values, r_squ):
                        sources_to_remove.append(source_number)
                    else: # Linear motion in right ascension
                        # Linear motion in declination?
                        if not motion_is_linear(epoch, dec, group['ERRB_WORLD'].values, r_squ):
                            if not motion_is_linear_in_ecliptic_coordinates(epoch, ra, dec, r_squ):
                                sources_to_remove.append(source_number)
                    continue # source has linear motion and is kept

                else: # outlier found in middle (jump in epochs)
                    outlier_index = outlier_in_middle

            # Found outlier in EPOCH
            subgroups = np.split(group[['EPOCH', 'ALPHA_J2000', 'DELTA_J2000',
                                        'ERRA_WORLD', 'ERRB_WORLD']], outlier_index + 1)

            for subgroup in subgroups:

                if len(subgroup) < min_detections: # if there aren't sufficient detections
                    sources.loc[subgroup.index, 'FLAGS_SSOS'] +=1  # Add outlier flag

                else:
                    epoch = subgroup['EPOCH'].values
                    ra = subgroup['ALPHA_J2000'].values
                    dec = subgroup['DELTA_J2000'].values
                    # Linear in right ascencsion?
                    if not motion_is_linear(epoch, ra, subgroup['ERRA_WORLD'].values, r_squ):
                        sources.loc[subgroup.index, 'FLAGS_SSOS'] +=1  # Add outlier flag
                    else: # Linear motion in right ascension
                        # Linear motion in declination?
                        if not motion_is_linear(epoch, dec, subgroup['ERRB_WORLD'].values, r_squ):
                            if not motion_is_linear_in_ecliptic_coordinates(epoch, ra, dec, r_squ):
                                sources.loc[subgroup.index, 'FLAGS_SSOS'] +=1  # Add outlier flag

            # FLAG_SSOS is only uneven if detection is outlier
            # If all source detections were flagged, remove the source
            if all(sources[sources.SOURCE_NUMBER == source_number]['FLAGS_SSOS'] % 2 == 1):
                sources_to_remove.append(source_number)

        else:
            # Linear in right ascencsion?
            if not motion_is_linear(epoch, ra, group['ERRA_WORLD'].values, r_squ):
                sources_to_remove.append(source_number)

            else: # Linear motion in right ascension
                # Linear motion in declination?
                if not motion_is_linear(epoch, dec, group['ERRB_WORLD'].values, r_squ):
                    if not motion_is_linear_in_ecliptic_coordinates(epoch, ra, dec, r_squ):
                        sources_to_remove.append(source_number)

    sources = sources[~sources.SOURCE_NUMBER.isin(sources_to_remove)]
    return sources


def compare_std_and_sigma_of_average(y, sigma):
    ''' Compares weighted average uncertainty to standard deviation '''
    weights = 1 / sigma**2
    weighted_average_uncertainty = 1 / np.sqrt(np.sum(weights))
    standard_deviation_of_data = np.std(y)
    ratio = weighted_average_uncertainty / standard_deviation_of_data

    return ratio


def constant_trail(sources, settings):
    '''
    Remove sources based on constant trail size

    input
    ------
    sources - pd.DataFrame, source data
    settings - dict, 'parameter': value

    return
    ------
    sources - pandas dataframe, sources data without the sources filtered in this step
    '''

    ratio = settings['RATIO']

    for source_number, group in sources.copy().groupby('SOURCE_NUMBER'):

        awin = group['AWIN_IMAGE']
        bwin = group['BWIN_IMAGE']

        for i, dimension in enumerate([awin, bwin]):

            sigma = group['ERRAWIN_IMAGE'] if i == 0 else group['ERRBWIN_IMAGE']

            fitted_ratio = compare_std_and_sigma_of_average(dimension, sigma)

            if fitted_ratio >= ratio and i == 0:
                pass
            elif fitted_ratio >= ratio and i == 1:
                pass
            else:
                sources = sources[sources.SOURCE_NUMBER != source_number]

    return sources


def bright_sources_catalog(sources, settings):
    '''
        Remove sources based on proximity to stars. Requires the HYG stellar
        catalog.

        input
        ------
        sources - pandas dataframe, table of source data
        settings - dict, 'parameter': value

        return
        ------
        sources - pandas dataframe, sources data without the sources filtered in this step
    '''

    # If the bright sources cat is the SCAMP reference cat, we have to find the filename
    if type(settings['BRIGHT_SOURCES_CAT']) == list:
        refout_catpath, astref_catalog = settings['BRIGHT_SOURCES_CAT']
        try:
            settings['BRIGHT_SOURCES_CAT'] = [file for file in os.listdir(refout_catpath)
                                                       if file.startswith(astref_catalog)][0]
        except IndexError:
            from ssos.core import PipelineSettingsException
            raise PipelineSettingsException('Could not find SCAMP reference catalogue %s\n' \
                                             'Run SCAMP to retrieve it to file.' % astref_catalog)
    try:
        with fits.open(settings['BRIGHT_SOURCES_CAT']) as ref_cat:
            bright_sources = Table(ref_cat[2].data).to_pandas()
    except OSError:
        try:
            bright_sources = pd.read_csv(settings['BRIGHT_SOURCES_CAT'])
        except ParserError:
            from ssos.core import PipelineSettingsException
            raise PipelineSettingsException('Could not read bright-sources catalogue %s ' % settings['BRIGHT_SOURCES_CAT'])

    bright_sources.rename(index=str, columns={'X_WORLD': 'RA', 'Y_WORLD': 'DEC'}, inplace=True)

    distance = settings['DISTANCE'] / 3600 # convert distance to degree
    lower_mag_limit, upper_mag_limt = [int(limit) for limit in settings['MAG_LIMITS'].split(',')]

    # Remove reference sources which are too faint or too bright
    bright_sources = bright_sources[(lower_mag_limit <= bright_sources.MAG) &
                                    (bright_sources.MAG <= upper_mag_limt)]

    # Remove bright_sources which are outside region of interest, given by the minimum and maximum
    # RA and Dec values of the source candiates plus the distance of the stellar regions
    upper_ra, lower_ra = (sources['ALPHA_J2000'].max() + distance,
                          sources['ALPHA_J2000'].min() - distance)
    upper_dec, lower_dec = (sources['DELTA_J2000'].max() + distance,
                            sources['DELTA_J2000'].min() - distance)

    bright_sources = bright_sources[(bright_sources['RA'] > lower_ra) &\
                                    (bright_sources['RA'] < upper_ra)]
    bright_sources = bright_sources[(bright_sources['DEC'] > lower_dec) &\
                                    (bright_sources['DEC'] < upper_dec)]

    for source_number, group in sources.copy().groupby('SOURCE_NUMBER'):

            mean_ra = np.mean(group['ALPHA_J2000'])
            mean_dec = np.mean(group['DELTA_J2000'])

            try:
                min_distance_to_star = min(np.sqrt((bright_sources['RA'] - mean_ra)**2 +
                                                   (bright_sources['DEC'] - mean_dec)**2).abs())

            except ValueError:  # reports "min() arg is an empty sequence"
                continue

            if min_distance_to_star < distance:
                sources = sources[sources.SOURCE_NUMBER != source_number]

    return sources


'''
    Extraction of cutouts of SSOs using SWarp,
    cross-matching with the SkyBoT database, and other analyses.

    Part of the SSO recovery pipeline.
'''

import os

from astropy.io import fits
from astropy.time import Time
import numpy as np
import pandas as pd


def extract_cutouts(sources, settings, log, paths, args):
    '''
        Extract cutouts of sources found in SCAMP catalogues using SWARP

        input
        ------
        sources - pd.DataFrame, table of source data
        settings - dict, dictionary of pipeline settings with PARAM:VALUE pairs
        log - logging object
        paths - dict, absolute paths of target directories
        args - command line argument parser

        return
        ------
        sources - pd.DataFrame, table of source data (unchanged)
    '''

    log.info('\nExtracting cutouts with SWARP..')

    image_dir    = args.fields[0]
    cutout_dir   = paths['cutouts']
    cutout_size  = settings['CUTOUT_SIZE']
    swarp_config = settings['SWARP_CONFIG']

    for index, row in sources.iterrows():
        img_ext = settings['SCI_EXTENSION'][row['EXTENSION'] - 1] if settings['SCI_EXTENSION'] else row['EXTENSION'] - 1
        image_file = os.path.join(image_dir, row['IMAGE_FILENAME']) + '[%i]' % img_ext

        cutout_filename = os.path.join(cutout_dir, '{:.0f}_{:02d}.fits'.format(row['SOURCE_NUMBER'],
                                                                               row['CATALOG_NUMBER']))

        if not args.swarp and os.path.isfile(cutout_filename):
            log.debug('Cutout %s exists, skipping..' % cutout_filename)
            continue

        swarp_args = {
                    'file': image_file,
                    'config': swarp_config,
                    'overwrite_params': {
                        'WEIGHTOUT_NAME': os.path.join(paths['tmp'], 'tmp_weight'),
                        'CENTER'        : '{:.5f},{:.5f}'.format(row['ALPHA_J2000'], row['DELTA_J2000']),
                        'IMAGEOUT_NAME' : '{:s}'.format(cutout_filename),
                        'IMAGE_SIZE'    : str(cutout_size),
                        'COPY_KEYWORDS' : 'OBJECT,RA,DEC,FILTER,DATE-OBS'},
                    }
        # ------
        # Build command and execute SWARP
        cmd = ' '.join(['swarp', swarp_args['file'], '-c', swarp_args['config']])
        for param, value in swarp_args['overwrite_params'].items():
            cmd += ' '.join([' -' + param, value])

        log.debug('\nExecuting SWARP command:\n%s' % cmd)
        if log.level > 10:  # Shown SWARP warnings only when debugging
            cmd += ' >/dev/null 2>&1'
        os.system(cmd)

    log.info('\tDone.\n')
    return sources


def compute_aperture_magnitudes(sources, settings, log, paths, args):
    '''
    Executes SExtractor in dual image mode and computes fixed aperture magnitudes

    input
    ------
    sources - pd.DataFrame, table of source data
    settings - dict, dictionary of pipeline settings with PARAM:VALUE pairs
    log - logging object
    paths - dict, absolute paths of target directories
    args - command line argument parser

    return
    ------
    sources - pd.DataFrame, table of source data with new magnitudes
    '''

    log.info('Computing fixed aperture magnitudes with SExtractor..')

    cutout_dir = paths['cutouts']
    cutouts = [os.path.join(cutout_dir, cutout) for cutout in os.listdir(cutout_dir)
               if cutout.endswith('.fits')]
    reference_filter = settings['REFERENCE_FILTER'].split(',')

    for number, source in sources.groupby('SOURCE_NUMBER'):

        # Select detection image based on reference filter
        filter_found = False
        for filter_ in reference_filter:
            try:
                reference_detection = source.loc[source['FILTER'] == filter_].iloc[0]
                filter_found = True
            except IndexError:
                continue
            if filter_found:
                sources.loc[(sources.SOURCE_NUMBER == reference_detection.SOURCE_NUMBER) &
                            (sources.CATALOG_NUMBER == reference_detection.CATALOG_NUMBER), 'FLAGS_SSOS'] += 2
                break
        else:
            from ssos.core import PipelineSettingsException
            raise PipelineSettingsException('No detection of candidate %i found in reference filter %s.'\
                                            % (number, reference_filter))

        detection_image = os.path.join(cutout_dir, '_'.join([str(number),
                                                             str(reference_detection['CATALOG_NUMBER'])])
                                                             + '.fits')

        # ------
        # Call SExtractor in dual image mode
        for image in cutouts:

            if str(number) not in image:
                continue

            cat = os.path.join(paths['cutouts'], os.path.basename(image.replace('.fits', '.cat')))

            if image == detection_image:
                detection_cat = cat

            sex_args = {
                'file': ','.join([detection_image, image]),
                'config': settings['SEX_CONFIG'],
                'overwrite_params': {
                    'PARAMETERS_NAME': settings['SEX_PARAMS'],
                    'CATALOG_NAME'   : cat,
                    'WEIGHT_TYPE'    : 'BACKGROUND',
                    'FILTER_NAME'    : settings['SEX_FILTER'],
                    'STARNNW_NAME'   : settings['SEX_NNW'],
                },
            }

            # ------
            # Exectue SExtractor
            cmd = ' '.join(['sex', sex_args['file'], '-c', sex_args['config']])
            for param, value in sex_args['overwrite_params'].items():
                cmd += ' '.join([' -' + param, value])
            os.system(cmd)

        # Identify the SSO in the detection cat to track it among the other ones
        cutout_size = settings['CUTOUT_SIZE']

        with fits.open(detection_cat) as detection_cat:

            # ------
            # Find the source closes to the center coordinate
            x_images = detection_cat[2].data.field('XWIN_IMAGE')
            y_images = detection_cat[2].data.field('YWIN_IMAGE')

            # Get 2d distances of sources to image center
            distances = [np.linalg.norm(np.array([cutout_size, cutout_size]) / 2 - np.array([x_img, y_img]))
                         for x_img, y_img in zip(x_images, y_images)]

            # Find index of source closest to center -> SSO candidate
            center_source_ind = np.where(distances == np.min(distances))[0]

        for ind, detection in source.iterrows():

            catalog = os.path.join(cutout_dir, '_'.join([str(detection.SOURCE_NUMBER),
                                                         str(detection.CATALOG_NUMBER)]) + '.cat')

            with fits.open(catalog) as cat:
                for prop in ['MAG', 'FLUX']:
                    sources.loc[ind, prop + '_CI'] = cat[2].data.field(prop + '_AUTO')[center_source_ind]
                    sources.loc[ind, prop + 'ERR_CI'] = cat[2].data.field(prop + 'ERR_AUTO')[center_source_ind]

    log.info('\tDone.\n')

    return sources


def _cross_match(source, skybot, radius):
    '''
        Cross-matching is applied to single rows in database.  Checks for matches
        and adds SkyBoT source parameters to database if a match is found.

        input
        ------
        source - pd.Series, single source detection parameters
        skybot - pd.DataFrame, parameters of all SkyBoT sources in exposure
        radius - float, crossmatch radius in arcsec

        returns
        source - pd.Series, single source detection parameters with SkyBot parameters added
    '''

    # Find SkyBoT source that is closest in angular separation to source
    distances_to_source = abs(np.sqrt((skybot['RA'] - source['ALPHA_J2000'])**2 +\
                                      (skybot['DEC'] - source['DELTA_J2000'])**2))
    if distances_to_source.min() <= radius:
        skybot_match = skybot.iloc[distances_to_source.values.argmin()]

        source['SKYBOT']         = True
        source['SKYBOT_NUMBER']  = skybot_match['#Num']
        source['SKYBOT_NAME']    = skybot_match['Name']
        source['SKYBOT_MAG']     = skybot_match['Mv']
        source['SKYBOT_ALPHA']   = skybot_match['RA']
        source['SKYBOT_DELTA']   = skybot_match['DEC']
        source['SKYBOT_PMALPHA'] = skybot_match['dRA(arcsec/h)']
        source['SKYBOT_PMDELTA'] = skybot_match['dDEC(arcsec/h)']
        source['SKYBOT_CLASS']   = skybot_match['Class']
    else:
        source['SKYBOT'] = False

    return source


def _compute_pm_difference_angle(row):
    # returns the angle between the skybot and source proper motion vectors
    skybot_pm = np.array([row.SKYBOT_PMALPHA, row.SKYBOT_PMDELTA])

    try:
        skybot_pm /= np.linalg.norm(skybot_pm)

    except ValueError: # no skybot match
        return 100.    # should never be the largest angle

    source_pm = np.array([row.PMALPHA_J2000, row.PMDELTA_J2000])
    source_pm /= np.linalg.norm(source_pm)

    return np.arccos(np.clip(np.dot(source_pm, skybot_pm), -1.0, 1.0))


def crossmatch_skybot(sources, settings, log, paths, args):
    '''
    Query SkyBoT for each epoch and cross-match the expected SSOs
    with the sources in the database.  Requires working internet connection.

    input
    ------
    sources - pd.DataFrame, table of source data
    settings - dict, dictionary of pipeline settings with PARAM:VALUE pairs
    log - logging object
    paths - dict, absolute paths of target directories
    args - command line argument parser

    return
    ------
    sources - pd.DataFrame, table of source data with SkyBoT matches added
    '''
    log.info('\nCross-matching SSO candidates with SkyBoT..')

    skybot = pd.read_csv(os.path.join(paths['skybot'], 'skybot_all.csv'))
    crossmatch_radius = settings['CROSSMATCH_RADIUS'] / 3600 # convert from arcsecond to degrees

    skybot_columns = ['SKYBOT', 'SKYBOT_NAME', 'SKYBOT_NUMBER', 'SKYBOT_CLASS', 'SKYBOT_MAG',
                      'SKYBOT_ALPHA', 'SKYBOT_DELTA', 'SKYBOT_PMALPHA', 'SKYBOT_PMDELTA']

    # The default value for the SkyBoT columns are empty strings
    sources = pd.concat([sources, pd.DataFrame('', sources.index,
                        skybot_columns)], axis=1)


    # Cross-match for each mid-exposure epoch
    for mid_exp_mjd, detections in sources.groupby('MID_EXPOSURE_MJD'):

        skybot_detections = skybot[skybot.EPOCH == Time(mid_exp_mjd, format='mjd').isot]
        sources.loc[detections.index, skybot_columns] = detections.apply(_cross_match, skybot=skybot_detections,
                                                                                       radius=crossmatch_radius, axis=1)


    # For each SSO, check if multiple matches were found
    # If so, choose the match which is most co-linear with the measured proper motion (minimize the angles)
    for number, source in sources.groupby('SOURCE_NUMBER'):

        if len(set(source.SKYBOT_NAME)) > 1:
            # if one of the matches is NaN, remove it and see if there's only one left
            names = set([name for name in source.SKYBOT_NAME if name != ''])

            if len(names) == 1: # some detections were not matched, add the designation to their columns

                numbers = [numb for numb in source.SKYBOT_NUMBER if numb != '']
                sources.loc[source.index, 'SKYBOT_NAME'] = list(names)[0]
                sources.loc[source.index, 'SKYBOT_NUMBER'] = list(numbers)[0]

            else: # compute the best match by minimizing the angular difference in proper motion

                pm_difference = source.apply(_compute_pm_difference_angle, axis=1)
                best_match = pm_difference.idxmin()

                best_match_name = sources.loc[best_match, 'SKYBOT_NAME']
                best_match_number = sources.loc[best_match, 'SKYBOT_NUMBER']

                sources.loc[source[source.SKYBOT_NAME != best_match_name].index,
                                  skybot_columns] = [best_match_name, best_match_number,
                                                                         '', '', '', '', '', '']

    log.info(' %i matched.\n' % len(set(sources[sources.SKYBOT]['SOURCE_NUMBER'])))
    return sources


#!/usr/bin/env python3
'''
  Pipeline to search for Solar System objects in wide-field imaging surveys
  Information on the project can be found in arXiv:1711.02780

  For questions, contact max.mahlke (at) cab.inta-csic.es

  Max Mahlke, August 2018
'''

import os
import sys
import time

if len(sys.argv) == 1:
    from ssos import GREETING
    print(GREETING)
    sys.exit()

if sys.argv[1] in ['-d', '--default']:
    path_to_module = os.path.dirname(__file__)
    os.system('cp -i -r {%s,%s} .' % (os.path.join(path_to_module, 'semp'),
                                   os.path.join(path_to_module, 'default.ssos')))
    sys.exit()

elif sys.argv[1] in ['-i', '--inspect']:
    from ssos.inspect import inspectCutouts
    inspectCutouts(sys.argv[2:])
    sys.exit()


from ssos.core import Pipeline


def main():
    # ========
    # Start the pipeline.  This step initializes the log, the target directory,
    # evaluates and checks the pipeline settings and verifies the input images.
    # ========
    pipeline = Pipeline()

    # ------
    # Run SExtractor on images
    pipeline.log.info('\nRunning SExtractor..\t')
    pipeline.run_SExtractor()

    # ------
    # Run SCAMP on SExtractor catalogues
    if pipeline.settings['REMOVE_REF_SOURCES']:

        pipeline.run_SCAMP(crossid_radius=1, full_name='full_stars.cat',
                           merged_name='merged_stars.cat', keep_refcat=True,
                           adjust_SExtractor_and_aheader=True)
        pipeline.run_SCAMP(crossid_radius=pipeline.args.CROSSID_RADIUS, solve_astronomy=False, pattern_matching=False)

    else:
        pipeline.run_SCAMP(crossid_radius=pipeline.args.CROSSID_RADIUS)

    # ========
    # Catalogue creation done
    # ========
    pipeline.log.info('\n --- Starting Filter pipeline ---\n\n')
    pipeline.log.info('%s %i\n' % ('All Sources'.ljust(20), pipeline.number_of_sources()))

    # Call pipeline filter steps
    for step in pipeline.steps:
        pipeline.execute_filter(step)
        pipeline.log.info('%s %i \n' % (step.ljust(20), pipeline.number_of_sources()))

    if not pipeline.added_proper_motion:
        pipeline.add_proper_motion()

    if not pipeline.added_SExtractor_data:
        pipeline.add_SExtractor_data()

    # ========
    # FILTERING COMPLETE
    # ========

    # ------
    # Optional analyses

    pipeline.add_image_metadata()

    for step in pipeline.analysis_steps:
        pipeline.execute_analysis(step)

    # ------
    # Rename columns, remove unnecessary data, save to file
    pipeline.save_and_cleanup()

    pipeline.log.info('\t|\t'.join(
                     ['\nAll done!',
                      '%i SSO candidates found' % pipeline.number_of_sources(),
                      'The analysis ran in %i seconds\n\n' % pipeline.run_time]))

    pipeline.log.info('Output File: %s\nLog File: %s\n\n' %
                     (os.path.join(pipeline.paths['cats'],
                                   'ssos_{:s}.csv'.format(time.strftime('%Y%m%d%H%M%S',
                                                          pipeline.start_time))),
                      os.path.join(pipeline.paths['logs'], pipeline.log_file)))


if __name__ == '__main__':
    main()



#!/usr/bin/env python3
from setuptools import setup, find_packages
from os import path

with open('README.md', 'r') as rm:
    long_description = rm.read()

setup(
    name = 'ssos',
    version = '1.2.3',
    description = 'The ssos Pipeline - Identiciation of Solar System Objects in astronomical images',
    url = 'https://github.com/maxmahlke/ssos',
    author = 'Max Mahlke',
    author_email = 'max.mahlke@cab.inta-csic.es',
    long_description = long_description,
    long_description_content_type = 'text/markdown',
    packages=find_packages(exclude = ['contrib', 'docs', 'tests']),
    keywords = ['astronomy astrophysics solar-system data pipeline'],
    setup_requires = [
                      'astropy',
                      'pandas>=0.23.0',
                      'numpy',
                      'scipy',
                      'statsmodels>=0.9.0'
                      ],
    install_requires = [
                        'astropy',
                        'pandas>=0.23.0',
                        'numpy',
                        'scipy',
                        'statsmodels>=0.9.0'
                        ],
    entry_points = {
                    'console_scripts': ['ssos = ssos.__main__:main']
                   },
    include_package_data=True,
    python_requires = '>=3'
    )



import os

from astropy.io import fits
from astropy.visualization import simple_norm, ZScaleInterval
from astropy.wcs import WCS
import matplotlib.animation as animation
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd

import warnings
warnings.filterwarnings('ignore')

def press(event, ssos, index, source_number):

    ssos.loc[index, 'INSPECTED'] = True

    if event.key in ['a', 'right']:
        ssos.loc[index, 'ASTEROID'] = True
        print('%i: Asteroid' % source_number)

    if event.key in ['u', 'up']:
        ssos.loc[index, 'UNKNOWN'] = True
        print('%i: UFO' % source_number)

    if event.key in ['d', 'left']:
        ssos.loc[index, 'ASTEROID'] = False
        print('%i: Discarded' % source_number)

    plt.close()



def inspectCutouts(ana_dirs):


    for ana_dir in ana_dirs:

        try:
            sso_path = os.path.join(sorted([os.path.join(ana_dir, 'cats/', csv) for csv in
                                          os.listdir(os.path.join(ana_dir, 'cats/')) if csv.startswith('ssos')
                                          and csv.endswith('csv')])[-1])
        except IndexError:
            print('\nNo SSO candidates database found in %s' % ana_dir)
            continue

        # Open SSO candidates data
        ssos = pd.read_csv(sso_path)
        print('Looking at %i candidates in %s' % (len(set(ssos.SOURCE_NUMBER)), ana_dir))


        if 'INSPECTED' not in ssos.columns:
            ssos['INSPECTED'] = False
            ssos['ASTEROID'] = False
            ssos['UNKNOWN'] = False


        for sn, sso in ssos.groupby('SOURCE_NUMBER'):

            # Get cutout filepaths
            cutouts = [os.path.join(ana_dir, 'cutouts', str(sn) + '_{:02d}'.format(cat) + '.fits')
                       for cat in  sso.CATALOG_NUMBER]

            fig = plt.figure()
            fig.canvas.mpl_disconnect(fig.canvas.manager.key_press_handler_id)
            fig.canvas.mpl_connect('key_press_event', lambda event: press(event, ssos, sso.index, sn))

            ims = []

            image = fits.open(cutouts[0])[0]
            wcs = WCS(image.header, fobj=image)
            ax = plt.subplot(projection=wcs)

            for i, cutout in enumerate(cutouts):

                try:
                    image = fits.open(cutout)[0]
                except OSError:
                    print('Could not find %s in cutouts directory. Was EXTRACT_CUTOUTS set to True?' % cutout)
                    continue

                data = image.data
                norm = simple_norm(data, 'linear')
                interval = ZScaleInterval()

                # When data is zero we get a large contrast,
                # set to background level instead
                try:
                    data[data == 0] = np.min(data[data != 0])
                except ValueError: # complete frame zero
                    pass

                im = ax.imshow(data, cmap='gray',
                                      vmin=min(interval.get_limits(data)),
                                      vmax=max(interval.get_limits(data)),
                                      norm=norm, animated=True)
                if type(sso.SKYBOT_NAME.values[0]) != float:
                    ax.text(0, 1.05, sso.SKYBOT_NAME.values[0],
                       verticalalignment='center', transform=ax.transAxes)

                ims.append([im])

            ani = animation.ArtistAnimation(fig, ims, interval=150, blit=True)
            plt.show()

        print('Found %i asteroids, %i unknown objects, and discarded %i candidates.' %
              (len(set(ssos[ssos['ASTEROID']]['SOURCE_NUMBER'])), len(set(ssos[ssos['UNKNOWN']]['SOURCE_NUMBER'])), len(set(ssos[~ssos['ASTEROID']]['SOURCE_NUMBER']))))
        ssos.to_csv(sso_path, index=False)
